﻿Imports System.Net

Public Class Form1
    Private idHry As Integer
    Private pocetTahu As Integer
    Private mapa As List(Of List(Of IPolicko))
    Private ja As New Hrac()
    Private protivnik As New Nepritel()
    Private lastPaint
    Private pocetTahuNaZacatku As Integer = -1

    ''' <summary>
    ''' Odstartuje program.
    ''' </summary>
    ''' <remarks>Využívá HTTP</remarks>
    Private Sub btnStart_Click(sender As System.Object, e As System.EventArgs) Handles btnStart.Click
        LoadStandartOutput(command("start"))
        btnStart.Text = "Restart"
    End Sub

    ''' <summary>
    ''' Odešle požadavek GET a vrátí odpověď
    ''' </summary>
    ''' <param name="address">adresa požadvku</param>
    ''' <returns>odpověď serveru</returns>
    ''' <remarks>Adresa bude od začátku doplněna o server načtený z GUI</remarks>
    Private Function command(address) As String
        Dim wc As New WebClient()
        Return wc.DownloadString(String.Format("{0}/{1}", txtServer.Text, address))
    End Function

    ''' <summary>
    ''' vykreslí mapu
    ''' </summary>
    ''' <param name="lines">generické kolekce s daty</param>
    Private Sub printMap(lines As List(Of String))
        lastPaint = lines ' Uložení dat v případě ztráty (např při obnovení okna z minimalizace atd)
        Dim g = PictureBox1.CreateGraphics()

        ' znovunačtení mapy
        mapa = New List(Of List(Of IPolicko))
        For y = 0 To 9
            mapa.Add(New List(Of IPolicko))

            For x = 0 To 9
                Dim char1 = lines(y).Trim()(x * 2)
                Dim char2 = lines(y).Trim()(x * 2 + 1)
                Dim sequence = char1 & char2

                Select Case sequence
                    Case "++"
                        mapa(y).Add(New MojeZakladna())
                    Case "--"
                        mapa(y).Add(New TvojeZakladna())
                    Case ".."
                        mapa(y).Add(New VolnePolicko())
                    Case "XX"
                        mapa(y).Add(New Zed())
                    Case Else
                        mapa(y).Add(New Poklad(sequence))
                End Select

                ' vykreslení
                mapa(y)(x).NakresliSe(x, y, g)
            Next
        Next
        ' vykreslování hráčů
        ja.NakresliSe(g)
        If Not IsNothing(protivnik) Then
            protivnik.NakresliSe(g)
        End If
    End Sub

    ''' <summary>
    ''' Obnovení vykresleného obsahu
    ''' </summary>
    ''' <remarks>využívá dat uložených v lastPaint</remarks>
    Private Sub Form1_Paint(sender As Object, e As System.Windows.Forms.PaintEventArgs) Handles Me.Paint
        If Not IsNothing(lastPaint) Then
            printMap(lastPaint)
        End If
    End Sub

    ''' <summary>
    ''' Obsluha tlačítek pro pohyb. 
    ''' </summary>
    ''' <remarks>Příkaz je uložen ve vlastnosti Tag těchto tlačítek. Díky toho stačí jedna obluha pro 4 tlačítka</remarks>
    Private Sub btnPohyb_Click(sender As Button, e As System.EventArgs) Handles btnUp.Click, btnRight.Click, btnLeft.Click, btnDown.Click
        Dim output = command(String.Format("tah?id={0}&pohyb={1}", idHry, sender.Tag))
        LoadStandartOutput(output)
    End Sub

    ''' <summary>
    ''' Naparsuje standardní vstup
    ''' </summary>
    ''' <param name="output">výstup serveru</param>
    Private Sub LoadStandartOutput(output As String)
        Dim lines As Object = output.Split(New Char() {vbCrLf}, System.StringSplitOptions.RemoveEmptyEntries).ToList()
        Try ' V případě nestandartního vstupu se následující řádek nepovede
            idHry = lines(0)
        Catch ex As Exception ' vyresetování stavu
            MessageBox.Show(lines(0))
            idHry = Nothing
            mapa.Clear()
            pocetTahu = 0
            ja = New Hrac()
            protivnik = New Nepritel()
            lastPaint = Nothing
            validButtons()
            pocetTahuNaZacatku = -1
            btnStart.Text = "Start"
            lblTahy.Text = "Tahy došly"
            Return
        End Try
        pocetTahu = lines(1)
        If Me.pocetTahuNaZacatku = -1 Then
            pocetTahuNaZacatku = pocetTahu
        End If
        ' odstranění použitých aby se nedostaly do vstupu pro vykreslování
        lines.RemoveAt(0)
        lines.RemoveAt(0)

        ' data hráčů
        Dim lastVals = lines(10)
        ja.nactiData(lastVals)
        Try ' druhý hráš nemusí existovat
            lastVals = lines(11)
            protivnik.nactiData(lastVals)
        Catch ex As Exception
            protivnik = Nothing
        End Try

        ' odstranění dat hráčů pro vykreslováíání
        lines.RemoveAt(lines.Count - 1)
        lines.RemoveAt(lines.Count - 1)
        lines.RemoveAt(lines.Count - 1)

        ' výstup
        lblNeses.Text = String.Format("Neseš = {0}", ja.poklad)
        lblScore.Text = String.Format("Score = {0}", ja.score)
        lblTahy.Text = String.Format("Tahů {0}/{1}", pocetTahu, pocetTahuNaZacatku)

        printMap(lines)
        validButtons() ' zablokování zbytečných tlačítek
    End Sub

    ''' <summary>
    ''' Zab¨lokuje zbytečná tlačítka
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub validButtons()
        ' ty cathce jsou tu kvůli úpřetečení mimo mapu, kde index v mapě nebude existovat
        Try
            btnUp.Enabled = mapa(ja.y - 1)(ja.x).MuzesJitNa()
        Catch ex As Exception
            btnUp.Enabled = False
        End Try

        Try
            btnDown.Enabled = mapa(ja.y + 1)(ja.x).MuzesJitNa()
        Catch ex As Exception
            btnDown.Enabled = False
        End Try

        Try
            btnLeft.Enabled = mapa(ja.y)(ja.x - 1).MuzesJitNa()
        Catch ex As Exception
            btnLeft.Enabled = False
        End Try

        Try
            btnRight.Enabled = mapa(ja.y)(ja.x + 1).MuzesJitNa()
        Catch ex As Exception
            btnRight.Enabled = False
        End Try
    End Sub

    ''' <summary>
    ''' Provede jeden krok autopilata
    ''' </summary>
    Private Sub btnAutopilot_Click(sender As System.Object, e As System.EventArgs) Handles btnAutopilot.Click
        Dim autopilotTah As Char = autopilot()
        For Each c In Me.Controls
            If c.Tag = autopilotTah Then
                btnPohyb_Click(c, Nothing)
            End If
        Next
    End Sub

    ''' <summary>
    ''' Najde vhodný cíl, kam má autopilot nejkratší cestou dorazit
    ''' </summary>
    ''' <returns>bod vhodného cíle</returns>
    Private Function cilAutopilota() As Point
        If ja.poklad > 0 Then ' pokud máme naložený poklad, cíl bude základna
            For y = 0 To 9
                For x = 0 To 9
                    If TypeOf mapa(y)(x) Is MojeZakladna Then
                        Return New Point(x, y)
                    End If
                Next
            Next
        End If

        Dim poklad As Poklad = Nothing
        Dim xo As Integer
        Dim yo As Integer
        Dim l As Integer

        For y = 0 To 9
            For x = 0 To 9
                ' podmínka vymění vhodný cíl pokud
                ' a) poklad ještě nebyl vybrán
                ' b) pokud je hodnota vybraného pokladu nížší než prochledávaný poklad
                ' c) pokud je hodnota vybraného a prohledávaného pokladu sice stejná, ale prohledávaný poklad je blíže.
                If (IsNothing(poklad) And TypeOf mapa(y)(x) Is Poklad) Or
                    (Not IsNothing(poklad) AndAlso TypeOf mapa(y)(x) Is Poklad AndAlso CType(mapa(y)(x), Poklad).hodnotaPokladu > poklad.hodnotaPokladu) Or
                    (Not IsNothing(poklad) AndAlso TypeOf mapa(y)(x) Is Poklad AndAlso CType(mapa(y)(x), Poklad).hodnotaPokladu = poklad.hodnotaPokladu AndAlso l > delkaTrasy(x, y)) Then
                    poklad = mapa(y)(x)
                    xo = x
                    yo = y
                    l = delkaTrasy(x, y)
                End If
            Next
        Next
        Return New Point(xo, yo)
    End Function

    ''' <summary>
    ''' Vlnovým algoritmem najde vybere nejvhodnější směr pohybu
    ''' </summary>
    ''' <returns>L|P|N|D</returns>
    Private Function autopilot() As Char
        Dim vlnovaMapa As New List(Of List(Of Integer))
        Dim cil = cilAutopilota()
        Dim start = New Point(ja.x, ja.y)

        Dim s As New Queue(Of Point)


        s.Enqueue(start)

        For y = 0 To 9
            vlnovaMapa.Add(New List(Of Integer))
            For x = 0 To 9
                vlnovaMapa(y).Add(-1)
            Next
        Next

        vlnovaMapa(ja.y)(ja.x) = 1

        While s.Count > 0
            Dim p = s.Dequeue()

            If vlnovaMapa(p.Y)(p.X) > 10 * 10 Then
                Continue While
            End If
            If (p.X = cil.X And p.Y = cil.Y) Then
                Exit While
            End If

            ' x + 1

            Try
                If mapa(p.Y)(p.X + 1).MuzesJitNa() And vlnovaMapa(p.Y)(p.X + 1) < vlnovaMapa(p.Y)(p.X) - 1 Then
                    vlnovaMapa(p.Y)(p.X + 1) = vlnovaMapa(p.Y)(p.X) + 1
                    s.Enqueue(New Point(p.X + 1, p.Y))
                End If
            Catch ex As Exception

            End Try

            ' x  - 1

            Try
                If mapa(p.Y)(p.X - 1).MuzesJitNa() And vlnovaMapa(p.Y)(p.X - 1) < vlnovaMapa(p.Y)(p.X) - 1 Then
                    vlnovaMapa(p.Y)(p.X - 1) = vlnovaMapa(p.Y)(p.X) + 1
                    s.Enqueue(New Point(p.X - 1, p.Y))
                End If
            Catch ex As Exception

            End Try

            ' y + 1

            Try
                If mapa(p.Y + 1)(p.X).MuzesJitNa() And vlnovaMapa(p.Y + 1)(p.X) < vlnovaMapa(p.Y)(p.X) - 1 Then
                    vlnovaMapa(p.Y + 1)(p.X) = vlnovaMapa(p.Y)(p.X) + 1
                    s.Enqueue(New Point(p.X, p.Y + 1))
                End If
            Catch ex As Exception

            End Try


            ' y - 1

            Try
                If mapa(p.Y - 1)(p.X).MuzesJitNa() And vlnovaMapa(p.Y - 1)(p.X) < vlnovaMapa(p.Y)(p.X) - 1 Then
                    vlnovaMapa(p.Y - 1)(p.X) = vlnovaMapa(p.Y)(p.X) + 1
                    s.Enqueue(New Point(p.X, p.Y - 1))
                End If
            Catch ex As Exception

            End Try

        End While

        Dim trasa As New List(Of Point)

        Dim pb As Point = cil
        While pb <> start
            trasa.Add(pb)
            Try
                If vlnovaMapa(pb.Y)(pb.X - 1) < vlnovaMapa(pb.Y)(pb.X) And vlnovaMapa(pb.Y)(pb.X - 1) > -1 Then
                    pb.X -= 1
                    Continue While
                End If
            Catch ex As Exception

            End Try
            Try
                If vlnovaMapa(pb.Y)(pb.X + 1) < vlnovaMapa(pb.Y)(pb.X) And vlnovaMapa(pb.Y)(pb.X + 1) > -1 Then
                    pb.X += 1
                    Continue While
                End If
            Catch ex As Exception

            End Try
            Try
                If vlnovaMapa(pb.Y - 1)(pb.X) < vlnovaMapa(pb.Y)(pb.X) And vlnovaMapa(pb.Y - 1)(pb.X) > -1 Then
                    pb.Y -= 1
                    Continue While
                End If
            Catch ex As Exception

            End Try
            Try
                If vlnovaMapa(pb.Y + 1)(pb.X) < vlnovaMapa(pb.Y)(pb.X) And vlnovaMapa(pb.Y + 1)(pb.X) > -1 Then
                    pb.Y += 1
                    Continue While
                End If
            Catch ex As Exception

            End Try
        End While
        trasa.Add(start)
        trasa.Reverse()

        If trasa(0).X <> trasa(1).X Then ' měnila se pozice X
            If trasa(0).X > trasa(1).X Then ' p
                Return "L"
            Else
                Return "P"
            End If
        Else ' měnila se pozice Y
            If trasa(0).Y > trasa(1).Y Then ' p
                Return "N"
            Else
                Return "D"
            End If
        End If
    End Function

    ''' <summary>
    ''' Zjistí délku trasy do cíle
    ''' </summary>
    ''' <param name="xi">X cíle</param>
    ''' <param name="yi">Y cíle</param>
    ''' <returns>vzdálenost od cíle [políčka]</returns>
    Private Function delkaTrasy(xi As Integer, yi As Integer) As Integer
        Dim vlnovaMapa As New List(Of List(Of Integer))
        Dim cil = New Point(xi, yi)
        Dim start = New Point(ja.x, ja.y)

        Dim s As New Queue(Of Point)


        s.Enqueue(start)

        For y = 0 To 9
            vlnovaMapa.Add(New List(Of Integer))
            For x = 0 To 9
                vlnovaMapa(y).Add(-1)
            Next
        Next

        vlnovaMapa(ja.y)(ja.x) = 1

        While s.Count > 0
            Dim p = s.Dequeue()

            If vlnovaMapa(p.Y)(p.X) > 10 * 10 Then
                Continue While
            End If
            If (p.X = cil.X And p.Y = cil.Y) Then
                Exit While
            End If

            ' x + 1

            Try
                If mapa(p.Y)(p.X + 1).MuzesJitNa() And vlnovaMapa(p.Y)(p.X + 1) < vlnovaMapa(p.Y)(p.X) - 1 Then
                    vlnovaMapa(p.Y)(p.X + 1) = vlnovaMapa(p.Y)(p.X) + 1
                    s.Enqueue(New Point(p.X + 1, p.Y))
                End If
            Catch ex As Exception

            End Try

            ' x  - 1

            Try
                If mapa(p.Y)(p.X - 1).MuzesJitNa() And vlnovaMapa(p.Y)(p.X - 1) < vlnovaMapa(p.Y)(p.X) - 1 Then
                    vlnovaMapa(p.Y)(p.X - 1) = vlnovaMapa(p.Y)(p.X) + 1
                    s.Enqueue(New Point(p.X - 1, p.Y))
                End If
            Catch ex As Exception

            End Try

            ' y + 1

            Try
                If mapa(p.Y + 1)(p.X).MuzesJitNa() And vlnovaMapa(p.Y + 1)(p.X) < vlnovaMapa(p.Y)(p.X) - 1 Then
                    vlnovaMapa(p.Y + 1)(p.X) = vlnovaMapa(p.Y)(p.X) + 1
                    s.Enqueue(New Point(p.X, p.Y + 1))
                End If
            Catch ex As Exception

            End Try


            ' y - 1

            Try
                If mapa(p.Y - 1)(p.X).MuzesJitNa() And vlnovaMapa(p.Y - 1)(p.X) < vlnovaMapa(p.Y)(p.X) - 1 Then
                    vlnovaMapa(p.Y - 1)(p.X) = vlnovaMapa(p.Y)(p.X) + 1
                    s.Enqueue(New Point(p.X, p.Y - 1))
                End If
            Catch ex As Exception

            End Try
        End While

        Dim trasa As New List(Of Point)

        Dim pb As Point = cil
        While pb <> start
            trasa.Add(pb)
            Try
                If vlnovaMapa(pb.Y)(pb.X - 1) < vlnovaMapa(pb.Y)(pb.X) And vlnovaMapa(pb.Y)(pb.X - 1) > -1 Then
                    pb.X -= 1
                    Continue While
                End If
            Catch ex As Exception

            End Try
            Try
                If vlnovaMapa(pb.Y)(pb.X + 1) < vlnovaMapa(pb.Y)(pb.X) And vlnovaMapa(pb.Y)(pb.X + 1) > -1 Then
                    pb.X += 1
                    Continue While
                End If
            Catch ex As Exception

            End Try
            Try
                If vlnovaMapa(pb.Y - 1)(pb.X) < vlnovaMapa(pb.Y)(pb.X) And vlnovaMapa(pb.Y - 1)(pb.X) > -1 Then
                    pb.Y -= 1
                    Continue While
                End If
            Catch ex As Exception

            End Try
            Try
                If vlnovaMapa(pb.Y + 1)(pb.X) < vlnovaMapa(pb.Y)(pb.X) And vlnovaMapa(pb.Y + 1)(pb.X) > -1 Then
                    pb.Y += 1
                    Continue While
                End If
            Catch ex As Exception

            End Try
        End While
        trasa.Add(start)
        Return trasa.Count
    End Function

    Private Sub autopilotAll_Click(sender As System.Object, e As System.EventArgs) Handles autopilotAll.Click
        While Not IsNothing(idHry)
            Try
                btnAutopilot_Click(Nothing, Nothing)
            Catch ex As Exception
                Exit While
            End Try
            Me.Refresh()
        End While
    End Sub
End Class
